# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr
import pytest

from swh.journal.serializers import kafka_to_value
from swh.journal.writer import get_journal_writer
from swh.model import swhids
from swh.model.tests import swh_model_data
from swh.scrubber.journal_checker import JournalChecker


def journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group):
    return dict(
        cls="kafka",
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        on_eof="stop",
    )


def journal_writer(kafka_server, kafka_prefix):
    return get_journal_writer(
        cls="kafka",
        brokers=[kafka_server],
        client_id="kafka_writer",
        prefix=kafka_prefix,
        anonymize=False,
    )


def test_no_corruption(scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group):
    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("directory", swh_model_data.DIRECTORIES)
    writer.write_additions("revision", swh_model_data.REVISIONS)
    writer.write_additions("release", swh_model_data.RELEASES)
    writer.write_additions("snapshot", swh_model_data.SNAPSHOTS)

    JournalChecker(
        db=scrubber_db,
        journal=journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group),
    ).run()

    assert list(scrubber_db.corrupt_object_iter()) == []


@pytest.mark.parametrize("corrupt_idx", range(len(swh_model_data.SNAPSHOTS)))
def test_corrupt_snapshot(
    scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group, corrupt_idx
):
    snapshots = list(swh_model_data.SNAPSHOTS)
    snapshots[corrupt_idx] = attr.evolve(snapshots[corrupt_idx], id=b"\x00" * 20)

    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("snapshot", snapshots)

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    JournalChecker(
        db=scrubber_db,
        journal=journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group),
    ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == swhids.CoreSWHID.from_string(
        "swh:1:snp:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].datastore.package == "journal"
    assert corrupt_objects[0].datastore.cls == "kafka"
    assert (
        corrupt_objects[0].datastore.instance
        == f"brokers='{kafka_server}' prefix='{kafka_prefix}'"
    )
    assert (
        before_date - datetime.timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + datetime.timedelta(seconds=5)
    )
    assert (
        kafka_to_value(corrupt_objects[0].object_) == snapshots[corrupt_idx].to_dict()
    )


def test_corrupt_snapshots(
    scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    snapshots = list(swh_model_data.SNAPSHOTS)
    for i in (0, 1):
        snapshots[i] = attr.evolve(snapshots[i], id=bytes([i]) * 20)

    writer = journal_writer(kafka_server, kafka_prefix)
    writer.write_additions("snapshot", snapshots)

    JournalChecker(
        db=scrubber_db,
        journal=journal_client_config(kafka_server, kafka_prefix, kafka_consumer_group),
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 2
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in [
            "swh:1:snp:0000000000000000000000000000000000000000",
            "swh:1:snp:0101010101010101010101010101010101010101",
        ]
    }



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from unittest.mock import patch

from psycopg2.errors import UniqueViolation
import pytest

from swh.model.swhids import ObjectType
from swh.scrubber.db import Datastore, ScrubberDb

from .conftest import NB_PARTITIONS, OBJECT_TYPE

DATE = datetime.datetime(2022, 10, 4, 12, 1, 23, tzinfo=datetime.timezone.utc)
ONE_MINUTE = datetime.timedelta(minutes=1)
ONE_DAY = datetime.timedelta(days=1)


def test_config_add(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    cfg_snp = scrubber_db.config_add("cfg snp", datastore, ObjectType.SNAPSHOT, 42)
    assert cfg_snp == 2

    cfg_snp2 = scrubber_db.config_add("cfg snp 2", datastore, ObjectType.SNAPSHOT, 43)
    assert cfg_snp2 == 3

    cfg_snp3 = scrubber_db.config_add(None, datastore, ObjectType.SNAPSHOT, 44)
    assert cfg_snp3 == 4
    assert (
        scrubber_db.config_get(cfg_snp3).name == "check_snapshot_44_storage_postgresql"
    )

    # XXX this is debatable: is there a good reason to allow 2 configs for the
    # same datastore and object type but different partition number, but not
    # for the same number of partitions?
    with pytest.raises(UniqueViolation):
        scrubber_db.config_add("cfg4", datastore, OBJECT_TYPE, NB_PARTITIONS)


def test_config_get(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    cfg2 = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    cfg3 = scrubber_db.config_add("cfg3", datastore, ObjectType.SNAPSHOT, 43)

    assert scrubber_db.config_get(cfg2)
    assert scrubber_db.config_get(cfg3)

    with pytest.raises(ValueError):
        scrubber_db.config_get(cfg3 + 1)


def test_checked_config_get_by_name(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    cfg2 = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    cfg3 = scrubber_db.config_add("cfg3", datastore, ObjectType.SNAPSHOT, 43)

    assert scrubber_db.config_get_by_name("cfg2") == cfg2
    assert scrubber_db.config_get_by_name("cfg3") == cfg3

    assert scrubber_db.config_get_by_name("unknown config") is None


def test_datastore_get(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    assert scrubber_db.datastore_get(1) == datastore
    with pytest.raises(ValueError):
        scrubber_db.datastore_get(42)


def test_checked_partition_insert(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        part_gen = scrubber_db.checked_partition_iter_next(config_id)
        partition_id = next(part_gen)
    scrubber_db.checked_partition_upsert(config_id, partition_id, DATE + ONE_MINUTE)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (partition_id, NB_PARTITIONS, DATE, DATE + ONE_MINUTE)
    ]


def test_checked_partition_insert_two(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(part_gen)
        scrubber_db.checked_partition_upsert(config_id, part_id, DATE + ONE_MINUTE)

        config_snp = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
        snp_part_gen = scrubber_db.checked_partition_iter_next(config_snp)
        snp_part_id = next(snp_part_gen)
        scrubber_db.checked_partition_upsert(config_snp, snp_part_id, DATE + ONE_MINUTE)

    assert set(scrubber_db.checked_partition_iter(config_id)) == {
        (part_id, NB_PARTITIONS, DATE, DATE + ONE_MINUTE),
    }
    assert set(scrubber_db.checked_partition_iter(config_snp)) == {
        (snp_part_id, 42, DATE, DATE + ONE_MINUTE),
    }


def test_checked_partition_get_next(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    config_snp = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    snp_part_gen = scrubber_db.checked_partition_iter_next(config_snp)
    dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)

    assert next(dir_part_gen) == 0
    assert next(snp_part_gen) == 0
    assert next(dir_part_gen) == 1
    assert next(dir_part_gen) == 2
    assert next(snp_part_gen) == 1
    assert next(snp_part_gen) == 2
    assert next(dir_part_gen) == 3

    date = datetime.datetime.now(tz=datetime.timezone.utc)
    scrubber_db.checked_partition_upsert(config_snp, 0, date)
    scrubber_db.checked_partition_upsert(config_id, 2, date)

    assert next(snp_part_gen) == 3
    assert next(dir_part_gen) == 4


def test_checked_partition_get_next_with_hole(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)

    # fill the checked_partition table
    list(zip(range(20), dir_part_gen))

    # one hole at a time
    for part_id in range(10):
        assert scrubber_db.checked_partition_reset(config_id, part_id)
        assert next(dir_part_gen) == part_id

    # a series of holes
    for part_id in range(0, 10, 2):
        assert scrubber_db.checked_partition_reset(config_id, part_id)

    for i in range(5):
        assert next(dir_part_gen) == 2 * i

    # all the holes are filled, next partition is 20
    assert next(dir_part_gen) == 20


def test_checked_partition_update(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(dir_part_gen)
    scrubber_db.checked_partition_upsert(config_id, part_id, DATE + ONE_MINUTE)

    date2 = DATE + 2 * ONE_MINUTE
    scrubber_db.checked_partition_upsert(config_id, part_id, date2)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (part_id, NB_PARTITIONS, DATE, date2)
    ]

    date3 = DATE - ONE_MINUTE
    scrubber_db.checked_partition_upsert(config_id, part_id, date3)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (part_id, NB_PARTITIONS, DATE, date2)  # newest date wins
    ]


def test_checked_partition_get(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(dir_part_gen)
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id) is None

    scrubber_db.checked_partition_upsert(config_id, part_id, DATE)

    assert scrubber_db.checked_partition_get_last_date(config_id, part_id) == DATE


def test_checked_partition_get_running(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    assert list(scrubber_db.checked_partition_get_running(config_id)) == []
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id1 = next(dir_part_gen)
        part_id2 = next(dir_part_gen)
        part_id3 = next(dir_part_gen)

    assert scrubber_db.checked_partition_get_last_date(config_id, part_id1) is None
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id2) is None
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id3) is None

    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id1, DATE),
        (part_id2, DATE),
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id2, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id1, DATE),
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id1, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id3, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == []



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all objects in a swh-storage instance and recomputes their checksums."""

import logging
from typing import Any, Dict, List

from swh.journal.client import get_journal_client
from swh.journal.serializers import kafka_to_value
from swh.model import model

from .db import Datastore, ScrubberDb

logger = logging.getLogger(__name__)


class JournalChecker:
    """Reads a chunk of a swh-storage database, recomputes checksums, and
    reports errors in a separate database."""

    _datastore = None

    def __init__(self, db: ScrubberDb, journal: Dict[str, Any]):
        self.db = db
        self.journal_client_config = journal
        self.journal_client = get_journal_client(
            **journal,
            # Remove default deserializer; so process_kafka_values() gets the message
            # verbatim so it can archive it with as few modifications a possible.
            value_deserializer=lambda obj_type, msg: msg,
        )

    def datastore_info(self) -> Datastore:
        """Returns a :class:`Datastore` instance representing the journal instance
        being checked."""
        if self._datastore is None:
            config = self.journal_client_config
            if config["cls"] == "kafka":
                self._datastore = Datastore(
                    package="journal",
                    cls="kafka",
                    instance=(
                        f"brokers={config['brokers']!r} prefix={config['prefix']!r}"
                    ),
                )
            else:
                raise NotImplementedError(
                    f"StorageChecker(journal_client={self.journal_client_config!r})"
                    f".datastore()"
                )
        return self._datastore

    def run(self):
        """Runs a journal client with the given configuration.
        This method does not return, unless otherwise configured (with ``stop_on_eof``).
        """
        self.journal_client.process(self.process_kafka_messages)

    def process_kafka_messages(self, all_messages: Dict[str, List[bytes]]):
        for (object_type, messages) in all_messages.items():
            cls = getattr(model, object_type.capitalize())
            for message in messages:
                object_ = cls.from_dict(kafka_to_value(message))
                real_id = object_.compute_hash()
                if object_.id != real_id:
                    self.db.corrupt_object_add(
                        object_.swhid(), self.datastore_info(), message
                    )



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial

import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.model.swhids import ObjectType
from swh.scrubber.db import Datastore, ScrubberDb

scrubber_postgresql_proc = factories.postgresql_proc(
    load=[partial(initialize_database_for_module, modname="scrubber", version=6)],
)

postgresql_scrubber = factories.postgresql("scrubber_postgresql_proc")

OBJECT_TYPE = ObjectType.DIRECTORY
PARTITION_ID = 2
NB_PARTITIONS = 64


@pytest.fixture
def datastore():
    return Datastore(package="storage", cls="postgresql", instance="service=swh-test")


@pytest.fixture
def scrubber_db(postgresql_scrubber):
    db = ScrubberDb(postgresql_scrubber)
    with db.conn.cursor() as cur:
        cur.execute("TRUNCATE TABLE corrupt_object")
        cur.execute("TRUNCATE TABLE datastore CASCADE")
    return db


@pytest.fixture
def config_id(scrubber_db, datastore):
    return scrubber_db.config_add(
        f"cfg_{OBJECT_TYPE}_{NB_PARTITIONS}", datastore, OBJECT_TYPE, NB_PARTITIONS
    )
